"""Package a 'docker' Target into a docker image.

Uses the information from a `docker` target in order to build a
Dockerfile & build context and then uses those to construct a docker
image.

Each dependent build target is processed to produce DockerComponents
which has both: files, to be included in the docker build context and
'commands' which will be inserted into the Dockerfile. Each Docker
component has an 'order' attribute which controls when it executes
relative to other docker components.

An implicit command is generated to copy all files in the docker
component into the built image Once all files have been merged into a
digest and the Dockerfile generated, we shell out to docker in order
to actually build the image.

The resulting image is then tagged if any tags were configured as part
of the build target definition.

Please see the ./pants help docker for more information on available
docker target fields, and see the documentation for sources.py,
python_requirements.py for specifics on how the DockerComponents are
generated
"""
import itertools
import logging
from io import StringIO
from typing import Iterable, List, Optional

import sendwave.pants_docker.utils as utils
from pants.backend.python.subsystems.setup import PythonSetup
from pants.backend.python.target_types import PythonRequirementTarget
from pants.core.goals.package import BuiltPackage, BuiltPackageArtifact
from pants.core.util_rules.system_binaries import BinaryPathRequest, BinaryPaths
from pants.engine.environment import Environment, EnvironmentRequest
from pants.engine.fs import (
    AddPrefix,
    CreateDigest,
    Digest,
    FileContent,
    MergeDigests,
    Snapshot,
)
from pants.engine.process import Process, ProcessCacheScope, ProcessResult
from pants.engine.rules import Get, MultiGet, collect_rules, rule
from pants.engine.target import TransitiveTargets, TransitiveTargetsRequest
from pants.engine.unions import UnionMembership
from sendwave.pants_docker.docker_component import (
    DockerComponent,
    DockerComponentFieldSet,
)
from sendwave.pants_docker.python_requirement import VirtualEnvRequest
from sendwave.pants_docker.subsystem import Docker
from sendwave.pants_docker.target import DockerPackageFieldSet

logger = logging.getLogger(__name__)


def _build_tags(
    target_name: str, tags: List[str], registry: Optional[str]
) -> List[str]:
    """Build a list of docker tags.

    Each tag in the passed in list will be formatted as:
    {registry/}target_name:tag in order to tag a docker image
    """
    tags = [f"{target_name}:{tag}" for tag in tags]
    tags.append(target_name)
    if not registry:
        return tags
    return [f"{registry}/{tag}" for tag in tags]


def _build_tag_argument_list(
    target_name: str, tags: List[str], registry: Optional[str]
) -> List[str]:
    """Build a list of docker tags and format them as CLI arguments.

    Takes the name of the build artifact, a list of tags and an
    optional docker registry and formats them as arguments to the
    docker command line program. Adding a "-t" before each formatted
    "registry/name:tag

    i.e. ('test_container, ['version]) -> ["-t", "test-container:version"]

    If the 'tags' list is empty no tags will be produced
    """
    tags = _build_tags(target_name, tags, registry)
    tags = itertools.chain(*(("-t", tag) for tag in tags))
    return list(tags)


def _create_dockerfile(
    base_image: str,
    workdir: Optional[str],
    setup: Iterable[str],
    commands: Iterable[str],
    init_command: Iterable[str],
) -> str:
    """Construct a Dockerfile and write it into a string variable.

    Uses commands explicitly specified in the docker target definition &
    generated from the Target's dependencies
    """
    dockerfile = StringIO()
    dockerfile.write("FROM {}\n".format(base_image))
    if workdir:
        dockerfile.write("WORKDIR {}\n".format(workdir))
    if setup:
        dockerfile.writelines(["RUN {}\n".format(line) for line in setup])
    dockerfile.writelines(commands)
    dockerfile.write("COPY application .\n")
    if init_command:
        cmd = "CMD [{}]\n".format(
            ",".join('"{}"'.format(c) for c in init_command)
        )
        dockerfile.write(cmd)
    return dockerfile.getvalue()


@rule()
async def package_into_image(
    field_set: DockerPackageFieldSet,
    union_membership: UnionMembership,
    setup: PythonSetup,
    docker: Docker,
) -> BuiltPackage:
    """Build a docker image from a 'docker' build target.

    Creates a build context & dockerfile from the build target & its
    dependencies. Then builds & tags that image. (see the module
    docstring for more information)
    """
    target_name = field_set.address.target_name
    transitive_targets = await Get(
        TransitiveTargets, TransitiveTargetsRequest([field_set.address])
    )

    component_list = []
    created_virtual_env = False
    logger.debug("Building Target %s", target_name)
    for field_set_type in union_membership[DockerComponentFieldSet]:
        for target in transitive_targets.dependencies:
            if (
                isinstance(target, PythonRequirementTarget)
                and not created_virtual_env
            ):
                # if there are any third party python dependencies
                # create & activate a virtual env in the image, this
                # will copy in a constraints file (which will be used
                # when installing any 3rd-party dependencies)
                component_list.append(
                    Get(
                        DockerComponent,
                        VirtualEnvRequest(
                            setup.enable_resolves, setup.requirement_constraints
                        ),
                    )
                )
                # we only want one virtual env per image
                created_virtual_env = True
            if field_set_type.is_applicable(target):
                logger.debug(
                    "Dependent Target %s applies to as component %s",
                    target.address,
                    field_set_type.__name__,
                )
                component_list.append(
                    Get(
                        DockerComponent,
                        DockerComponentFieldSet,
                        field_set_type.create(target),
                    )
                )

    components = await MultiGet(*component_list)

    source_digests = []
    run_commands = []
    components = sorted(components, key=lambda c: c.order)
    for component in components:
        if component.sources:
            source_digests.append(component.sources)
        run_commands.extend(component.commands)
    source_digest = await Get(Digest, MergeDigests(source_digests))
    application_digest = await Get(
        Digest, AddPrefix(source_digest, "application")
    )
    dockerfile_contents = _create_dockerfile(
        field_set.base_image.value,
        field_set.workdir.value,
        field_set.image_setup.value,
        run_commands,
        field_set.command.value,
    )
    logger.info("Constructed Dockerfile:\n{}".format(dockerfile_contents))
    dockerfile = await Get(
        Digest,
        CreateDigest(
            [FileContent("Dockerfile", dockerfile_contents.encode("utf-8"))]
        ),
    )
    # create docker build context of all merged files & fetch docker
    # connection enviornment variables
    # and the location of the docker process
    search_path = ["/bin", "/usr/bin", "/usr/local/bin", "$HOME/"]
    docker_context, docker_env, docker_paths = await MultiGet(
        Get(Digest, MergeDigests([dockerfile, application_digest])),
        Get(Environment, EnvironmentRequest(utils.DOCKER_ENV_VARS)),
        Get(
            BinaryPaths,
            BinaryPathRequest(
                binary_name="docker",
                search_path=search_path,
            ),
        ),
    )
    if not docker_paths.first_path:
        raise ValueError(
            "Unable to locate Docker binary on paths: %s", search_path
        )
    process_path = docker_paths.first_path.path
    # build an list of arguments of the form ["-t",
    # "registry/name:tag"] to pass to the docker executable
    tag_arguments = _build_tag_argument_list(
        target_name, field_set.tags.value or [], field_set.registry.value
    )
    # create the image
    process_args = [process_path, "build"]
    process_args.extend(tag_arguments)
    process_args.append(".")  # use current (sealed) directory as build context
    if docker.options.report_progress:
        process_args.append("--progress")
        process_args.append("plain")
    process_result = await Get(
        ProcessResult,
        Process(
            env=docker_env,
            argv=process_args,
            input_digest=docker_context,
            description=f"Creating Docker Image from {target_name}",
            cache_scope=ProcessCacheScope.PER_SESSION,
        ),
    )
    if docker.options.report_progress:
        logger.info(process_result.stdout.decode())
        logger.info(process_result.stderr.decode())
    output_digest = await Get(Snapshot, Digest, process_result.output_digest)
    return BuiltPackage(
        digest=process_result.output_digest,
        artifacts=([BuiltPackageArtifact(f, ()) for f in output_digest.files]),
    )


def rules():
    """Return the pants rules for this module."""
    return [
        *collect_rules(),
    ]



"""Register Sendwave pants-docker plugin rules with the pants build system."""
import sendwave.pants_docker.package as package
import sendwave.pants_docker.python_requirement as python_requirement
import sendwave.pants_docker.sources as sources
import sendwave.pants_docker.subsystem as subsystem
import sendwave.pants_docker.target as target


def rules():
    """Collect all pants rules in the plugin."""
    return [
        *subsystem.rules(),
        *package.rules(),
        *sources.rules(),
        *python_requirement.rules(),
        *target.rules(),
    ]


def target_types():
    """Collect the one new target type we've added."""
    return [target.Docker]



"""Configuration for the Sendwave pants-docker plugin."""
from pants.engine.rules import SubsystemRule
from pants.option.option_types import BoolOption
from pants.option.subsystem import Subsystem


class Docker(Subsystem):
    """Options for Docker Plugin.

    Allow configuring plugin behavior under the [docker] scope in your
    pants.toml, or by passing --docker-{option-name} to your command.
    Available Options:

    report-progress (boolean): if true, log the output of the docker
        build process.
    """

    options_scope = "sendwave-docker"
    help = "Options for Docker Build Process."

    report_progress = BoolOption(
        "--report-progress",
        default=False,
        help="If true: the plugin will report output of `docker build`",
    )


def rules():
    """Register Docker options as a SubsystemRule."""
    return [SubsystemRule(Docker)]



import logging
from dataclasses import dataclass
from typing import Any, Optional

from pants.backend.python.subsystems.repos import PythonRepos
from pants.backend.python.subsystems.setup import PythonSetup
from pants.backend.python.target_types import PythonRequirementsField
from pants.core.util_rules.system_binaries import BinaryPathRequest, BinaryPaths
from pants.engine.fs import Digest, GlobMatchErrorBehavior, PathGlobs
from pants.engine.rules import Get, collect_rules, rule
from pants.engine.target import FieldSet
from pants.engine.unions import UnionRule
from pants.option.global_options import BootstrapOptions
from sendwave.pants_docker.docker_component import (
    DockerComponent,
    DockerComponentFieldSet,
)

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class VirtualEnvRequest:
    enable_resolves: bool
    requirement_constraints: Optional[str]


@rule
async def create_virtual_env(
    resolve_request: VirtualEnvRequest,
) -> DockerComponent:
    assert (
        not resolve_request.enable_resolves
    ), "Pants lockfiles not yet supported"

    copy_command = []
    sources = None
    pip_upgrade_constraint = ""
    if constraint_file := resolve_request.requirement_constraints:

        sources = await Get(
            Digest,
            PathGlobs(
                [constraint_file],
                glob_match_error_behavior=GlobMatchErrorBehavior.error,
                description_of_origin="the option `requirement_constraints`",
            ),
        )
        pip_upgrade_constraint = " -c {}".format(constraint_file)
        copy_command.append("COPY application/{} .\n".format(constraint_file))
    return DockerComponent(
        commands=tuple(
            [
                *copy_command,
                "RUN python -m venv --upgrade /.virtual_env\n",
                "ENV PATH=/.virtual_env/bin:$PATH\n",
                "ENV VIRTUAL_ENV=/.virtual_env\n",
                "RUN python -m pip install --upgrade pip{}\n".format(
                    pip_upgrade_constraint
                ),
            ]
        ),
        sources=sources,
    )


@dataclass(frozen=True)
class PythonRequirementsFS(FieldSet):
    required_fields = (PythonRequirementsField,)
    requirements: PythonRequirementsField


@rule
async def get_requirements(
    field_set: PythonRequirementsFS, setup: PythonSetup, repos: PythonRepos
) -> DockerComponent:
    assert not setup.enable_resolves, "Pants lockfiles not yet supported"
    if repos.repos:
        links_args = " ".join(
            "--find-links {}".format(repo for repo in repos.repos)
        )
    else:
        links_args = ""
    num_indices = len(repos.indexes)
    if num_indices >= 1:
        index_args = f"--index-url {repos.indexes[0]} "
    if num_indices > 1:
        index_args = index_args + " ".join(
            [f"--extra-index-url {url}" for url in repos.indexes[1:]]
        )
    if num_indices == 0:
        index_args = " --no-index "
    constraint_arg = ""
    if setup.requirement_constraints:
        constraint_arg = f"--constraint {setup.requirement_constraints}"

    commands = tuple(
        "RUN python -m pip install {} {} {} {}\n".format(
            index_args, links_args, constraint_arg, lib
        )
        for lib in field_set.requirements.value
    )
    return DockerComponent(
        commands=commands,
        sources=None,
    )


def rules():
    return [
        UnionRule(DockerComponentFieldSet, PythonRequirementsFS),
        *collect_rules(),
    ]



from dataclasses import dataclass

import pants.core.goals.package
from pants.core.goals.package import (
    BuiltPackage,
    BuiltPackageArtifact,
    OutputPathField,
)
from pants.engine.target import (
    COMMON_TARGET_FIELDS,
    Dependencies,
    DependenciesRequest,
    DescriptionField,
    HydratedSources,
    HydrateSourcesRequest,
    StringField,
    StringSequenceField,
    Tags,
    Target,
    Targets,
    TransitiveTargets,
    TransitiveTargetsRequest,
)
from pants.engine.unions import UnionRule


class BaseImage(StringField):
    alias = "base_image"
    required = True
    help = "This is used to set the Base Image for all future pants build steps (e.g. python:3.8.8-slim-buster)"


class DockerIgnore(StringSequenceField):
    alias = "docker_ignore"
    required = False
    default = []
    help = "A list of directories to exclude from the docker build context, each entry should be a valid line in a .dockerignore file"


class ImageSetup(StringSequenceField):
    alias = "image_setup_commands"
    required = False
    default = []
    help = 'Commands to run in the image during the build process. Each will be evaluated as it\'s own process in the container and will create a new layer in the resulting image (e.g. ["apt-get update && apt-get upgrade --yes", "apt-get -y install gcc libpq-dev"],)'


class WorkDir(StringField):
    alias = "workdir"
    required = False
    default = "container"
    help = "The directory inside the container into which"


class Registry(StringField):
    alias = "registry"
    required = False
    help = "The registry of the resulting docker image"


class Tags(StringSequenceField):
    alias = "tags"
    default = []
    required = False
    help = 'A list of tags to apply to the resulting docker image (e.g. ["1.0.0", "main"]) '


class Command(StringSequenceField):
    alias = "command"
    default = []
    required = False
    help = "Command used to run the Docker container"


@dataclass(frozen=True)
class DockerPackageFieldSet(pants.core.goals.package.PackageFieldSet):
    alias = "docker_field_set"
    required_fields = (BaseImage,)

    base_image: BaseImage
    image_setup: ImageSetup
    ignore: DockerIgnore
    registry: Registry
    tags: Tags
    dependencies: Dependencies
    workdir: WorkDir
    command: Command
    output_path: OutputPathField


class Docker(Target):
    help = "A docker image that will contain all the source files of its transitive dependencies"
    alias = "docker"
    core_fields = (
        DescriptionField,
        Dependencies,
        BaseImage,
        ImageSetup,
        OutputPathField,
        WorkDir,
        Tags,
        Command,
    )


def rules():
    return [
        UnionRule(
            pants.core.goals.package.PackageFieldSet, DockerPackageFieldSet
        )
    ]




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: pants_plugins/sendwave/pants_docker:pants_docker

from setuptools import setup

setup(**{
    'author': 'Nathan Rosenbloom, Jean Cochrane',
    'author_email': 'engineering@sendwave.com',
    'description': 'Pants Plugin to automatically generate docker images from pants targets',
    'install_requires': (
        'pantsbuild.pants.testutil<2.14,>=2.13.0a0',
        'pantsbuild.pants<2.14,>=2.13.0a0',
    ),
    'name': 'sendwave-pants-docker',
    'namespace_packages': (
    ),
    'package_data': {
    },
    'packages': (
        'sendwave.pants_docker',
    ),
    'python_requires': '==3.9.*',
    'url': 'https://github.com/waveremit/pants-docker',
    'version': '1.1.0',
})
